/* Copyright 2003-2005 Charles Robert Simpson, Jr.
 *
 * This file is part of NETI@home.
 * 
 * NETI@home is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * NETI@home is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NETI@home; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *
 * Program: NETI@home
 *
 * Purpose: Check out the README file.
 *
 * Author: Charles Robert Simpson, Jr. (Robby) (neti@ece.gatech.edu)
 *
 * Comments: Much of this code was taken from Martin Roesch's Snort program.
 *           Gotta love open source and code reuse!  See Martin's comments
 *           below:
 * Comments: Ideas and code stolen liberally from Mike Borella's IP Grab
 *           program. Check out his stuff at http://www.borella.net.  I
 *           also have ripped some util functions from TCPdump, plus Mike's
 *           prog is derived from it as well.  All hail TCPdump....
 *
 */

/*  I N C L U D E S  **********************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <errno.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <sys/stat.h>
#include <syslog.h>
#ifndef WIN32
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif  /* !WIN32 */
#include "timersub.h"

#include "signal.h"
#include "debug.h"
#include "util.h"

#include <ctype.h>

#ifndef WIN32
#include <netdb.h>
#include <string>
#endif

#include <zlib.h>   // for compression (make sure to compile with -lz)
#include "netistats.h"
#include "locations.h"

#define DEST_PORT 557
#define UPDATEALERT_PORT 558
#define DEST_ADDR "neti.ece.gatech.edu"
#define MAX_SIZE 300
// data is reported if there is data and no data has been reported for FORCE_SEND_TIME seconds
// currently == 24 hours
#define FORCE_SEND_TIME 86400
// data is erased if it grows over MAX_MSG bytes
#define MAX_MSG 4000000

#ifndef WIN32
#define LOGFILE "/var/log/netilog"
#define CONFFILE "/etc/neti.conf"
#endif

/* exported variables *********************************************************/
PV pv;                  /* program vars */
int datalink;           /* the datalink value */
char *progname;         /* name of the program (from argv[0]) */
char **progargs;
char *username;
char *groupname;
unsigned long userid = 0;
unsigned long groupid = 0;
struct passwd *pw;
struct group *gr;
pcap_t *pd;             /* pcap handle */
#ifndef WIN32
int updatealerted = 0;
#endif
int logfilesize;
time_t lastsendtime;
time_t lastemailsendtime;
char *lastemail;
list<char *> leftover_Flows_uncom;
list<int> leftover_Flows_uncom_size;
struct Neti_Header_v2 leftover_nh;
int leftover_nh_valid = 0;
int privacylevel;
time_t netistarttime;
struct pcap_stat lastps;

#ifdef WIN32
SOCKET netimapfd;
SOCKADDR_IN netimap_addr;
HANDLE HProperties = NULL;
HANDLE HExit = NULL;
char conffilename[80];
char logfilename[80];
#else
int netimapfd;
struct sockaddr_in netimap_addr;
#endif

PacketCount pc;         /* packet count information */
u_int snaplen;


grinder_t grinder;
static void Restart();
int send_stats();
int send_email();
int writeNETILog(int origsize, int bufsize, struct Neti_Header_v2 nh, char *origbuf);
int getPrivacy();
int getLogfileSize();
char * getInterface();
int getGeolocation();
int getIdentifier();
int getUSZip();
char * getEmail();
#ifdef WIN32
string getWindowsSystemType();
#else
int getLatestVersion();
#endif

/* Signal handler declarations ************************************************/
static void SigTermHandler(int signal);
static void SigIntHandler(int signal);
static void SigQuitHandler(int signal);
static void SigHupHandler(int signal);
static void SigUsr1Handler(int signal);

/*
 *
 * Function: main(int, char *)
 *
 * Purpose:  Handle program entry and exit, call main prog sections
 *           This can handle both regular (command-line) style
 *           startup, as well as Win32 Service style startup.
 *
 * Arguments: See command line args in README file
 *
 * Returns: 0 => normal exit, 1 => exit on error
 *
 */
int main(int argc, char* argv[]) 
{
#if defined(WIN32) && defined(ENABLE_WIN32_SERVICE)
    /* Do some sanity checking, because some people seem to forget to
     * put spaces between their parameters
     */
    if( argc > 1 &&
        ( _stricmp(argv[1], (SERVICE_CMDLINE_PARAM SERVICE_INSTALL_CMDLINE_PARAM))==0   ||
          _stricmp(argv[1], (SERVICE_CMDLINE_PARAM SERVICE_UNINSTALL_CMDLINE_PARAM))==0 ||
          _stricmp(argv[1], (SERVICE_CMDLINE_PARAM SERVICE_SHOW_CMDLINE_PARAM))==0       ) )
    {
        FatalError("You must have a space after the '%s' command-line parameter\n",
                   SERVICE_CMDLINE_PARAM);
        exit(0);
    }

    /* If the first parameter is "/SERVICE", then start NETI@home as a Win32 service */
    if( argc>1 && _stricmp(argv[1],SERVICE_CMDLINE_PARAM)==0)
    {
        return NETIServiceMain(argc, argv);
    }
#endif /* WIN32 && ENABLE_WIN32_SERVICE */

    return NETIMain(argc,argv);
}




/*
 *
 * Function: NETIMain(int, char *)
 *
 * Purpose:  The real place that the program handles entry and exit.  Called
 *           called by main(), or by NETIServiceMain().
 *
 * Arguments: See command line args in README file
 *
 * Returns: 0 => normal exit, 1 => exit on error
 *
 */
int NETIMain(int argc, char *argv[])
{
    int netiver = -1;
#ifdef WIN32
    HKEY hKey;
    LONG lRet;
    DWORD dwBufLen = 80;
#endif

    lastemail = NULL;

#ifndef WIN32
    #if defined(LINUX) || defined(FREEBSD) || defined(OPENBSD) || defined(SOLARIS)
        sigset_t set;

        sigemptyset(&set);
        sigprocmask(SIG_SETMASK, &set, NULL);
    #else
        sigsetmask(0);
    #endif
#endif  /* !WIN32 */

    /*    malloc_options = "AX";*/

    /* make this prog behave nicely when signals come along */
    signal(SIGTERM, SigTermHandler);
    signal(SIGINT, SigIntHandler);
    signal(SIGQUIT, SigQuitHandler);
    signal(SIGHUP, SigHupHandler);
    signal(SIGUSR1, SigUsr1Handler);

    /*
     * set a global ptr to the program name so other functions can tell what
     * the program name is
     */
    progname = argv[0];
    progargs = argv;

    // Map Stuff
#ifdef WIN32
    SOCKADDR_IN my_addr;
#else
    struct sockaddr_in my_addr;
#endif
    // End Map Stuff

#ifdef WIN32

#ifdef _WINNT_
    // Security Attributes (Not sure if this is the absolute safest way of doing this)
    SECURITY_ATTRIBUTES sa_properties;
    SECURITY_ATTRIBUTES sa_exit;
    SECURITY_DESCRIPTOR sd_properties;
    SECURITY_DESCRIPTOR sd_exit;

    sa_properties.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa_exit.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa_properties.bInheritHandle = TRUE;
    sa_exit.bInheritHandle = TRUE;
    sa_properties.lpSecurityDescriptor = &sd_properties;
    sa_exit.lpSecurityDescriptor = &sd_exit;

    // Properties Event
    if(!InitializeSecurityDescriptor(&sd_properties, SECURITY_DESCRIPTOR_REVISION)) {
        ErrorMessage("Could Not Initialize Properties Security Descriptor\n");
    }
    else {
        // Add NULL DACL to the security descriptor (so that other processes are able to signal this event)
        if(!SetSecurityDescriptorDacl(&sd_properties, TRUE, (PACL)NULL, FALSE)) {
            ErrorMessage("Could Not Add Properties NULL DACL\n");
        }
        else {
            HProperties = ::CreateEvent(&sa_properties, FALSE, FALSE, "NetiUpdatePropertiesEvent");
            if(HProperties == NULL) {
                ErrorMessage("Could Not Create Properties Event\n");
            }
        }
    }

    // Exit Event
    if(!InitializeSecurityDescriptor(&sd_exit, SECURITY_DESCRIPTOR_REVISION)) {
        ErrorMessage("Could Not Initialize Exit Security Descriptor\n");
    }
    else {
        if(!SetSecurityDescriptorDacl(&sd_exit, TRUE, (PACL)NULL, FALSE)) {
            ErrorMessage("Could Not Add Exit NULL DACL\n");
        }
        else {
            HExit = ::CreateEvent(&sa_exit, FALSE, FALSE, "NetiExitEvent");
            if(HExit == NULL) {
                ErrorMessage("Could Not Create Exit Event\n");
            }
        }
    }

#else

    // Properties Event
    HProperties = ::CreateEvent(NULL, FALSE, FALSE, "NetiUpdatePropertiesEvent");
    if(HProperties == NULL) {
        ErrorMessage("Could Not Create Properties Event\n");
    }

    // Exit Event
    HExit = ::CreateEvent(NULL, FALSE, FALSE, "NetiExitEvent");
    if(HExit == NULL) {
        ErrorMessage("Could Not Create Exit Event\n");
    }

#endif
#endif

#ifdef WIN32
    // Start winsock
    if (!init_winsock())
        FatalError("Could not Initialize Winsock!\n");

    // Set conffilename
    if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("Software\\NETI@home"), 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS) {
        FatalError("Cannot access registry\n");
    }
    lRet = RegQueryValueEx(hKey, TEXT(""), NULL, NULL, (LPBYTE)conffilename, &dwBufLen);
    RegCloseKey(hKey);
    if(lRet != ERROR_SUCCESS) {
        FatalError("Error reading registry\n");
    }
    // Append "\\neti.conf" to conffilename
    conffilename[dwBufLen] = '\0';
    strncat(conffilename, "\\neti.conf", 11);

    // Set logfilename
    dwBufLen = 80;
    if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("Software\\NETI@home"), 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS) {
        FatalError("Cannot access registry\n");
    }
    lRet = RegQueryValueEx(hKey, TEXT(""), NULL, NULL, (LPBYTE)logfilename, &dwBufLen);
    RegCloseKey(hKey);
    if(lRet != ERROR_SUCCESS) {
        FatalError("Error reading registry\n");
    }
    // Append "\\netilog" to logfilename
    logfilename[dwBufLen] = '\0';
    strncat(logfilename, "\\netilog", 9);
#endif

    lastsendtime = time(NULL);
    lastemailsendtime = 0;  // So email is initially reported

    memset(&lastps, 0, sizeof(struct pcap_stat));

#ifndef WIN32
    if(!updatealerted) {
        netiver = getLatestVersion();
        if(netiver == -1) {
            updatealerted = 0;
        }
        else if(netiver > NETI_VERSION) {
            ErrorMessage("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
            printf("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
            updatealerted = 1;
        }
        else {
            updatealerted = 1;
        }
    }
#endif

    memset(&pv, 0, sizeof(PV));
    
    /* turn on checksum verification by default */
    pv.checksums_mode = DO_IP_CHECKSUMS | DO_TCP_CHECKSUMS |
                        DO_UDP_CHECKSUMS | DO_ICMP_CHECKSUMS | DO_IGMP_CHECKSUMS;

#if defined(WIN32) && defined(ENABLE_WIN32_SERVICE)
    /* initialize flags which control the Win32 service */
    pv.terminate_service_flag = 0;
    pv.pause_service_flag = 0;
#endif  /* WIN32 && ENABLE_WIN32_SERVICE */

    /* chew up the command line */
    ParseCmdLine(argc, argv);

    // Obtain max logfile size
	logfilesize = getLogfileSize();
	if(logfilesize <= 0) {
        if(!pv.readmode_flag) {
            ErrorMessage("Can't obtain Logfile Size from configuration file\n");
        }
		logfilesize = 1000000;
	}
    // Obtain privacy level
	privacylevel = getPrivacy();
	if(privacylevel == -1) {
        ErrorMessage("Can't obtain Privacy Setting from configuration file\n");
		exit(1);
	}

	// Map Stuff using UDP
    // Without these memsets, this code does not work on Mac OSX
    // props to http://www.macosx.com/forums/archive/index.php/t-44667.html
    memset(&my_addr, 0, sizeof(my_addr));
    memset(&netimap_addr, 0, sizeof(netimap_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons(0);
	netimap_addr.sin_family = AF_INET;
	netimap_addr.sin_port = htons((u_short)1557);
#ifdef WIN32
	my_addr.sin_addr.S_un.S_addr = htonl(INADDR_LOOPBACK);
	netimap_addr.sin_addr.S_un.S_addr = htonl(INADDR_LOOPBACK);
#else
	my_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	netimap_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
#endif
	netimapfd = socket(AF_INET, SOCK_DGRAM, 0);
#ifdef WIN32
	if(netimapfd == INVALID_SOCKET) {
        ErrorMessage("Map Socket failed\n");
		return -1;
	}
#else
	if(netimapfd < 0) {
        ErrorMessage("Map Socket failed\n");
		return -1;
	}
#endif
	if(bind(netimapfd, (sockaddr *)&my_addr, sizeof(my_addr)) != 0) {
        ErrorMessage("Map bind failed\n");
		return -1;
	}

    /* If we are running non-root, install a dummy handler instead. */
    if (userid != 0)
        signal(SIGHUP, SigCantHupHandler);
    
    /* set the default logging dir if not set yet */
    if(!pv.log_dir)
    {
        if(!(pv.log_dir = strdup(DEFAULT_LOG_DIR)))
            FatalError("Out of memory setting default log dir\n");
    }
    
    /*
     * if we're not reading packets from a file, open the network interface
     * for reading.. (interfaces are being initalized before the config file
     * is read, so some plugins would be able to start up properly.
     */
    if(!pv.readmode_flag)
    {
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Opening interface: %s\n", 
                    PRINT_INTERFACE(pv.interface)););
        /* open up our libpcap packet capture interface */
        OpenPcap();
    }
    else
    {
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Opening file: %s\n", 
                    pv.readfile););

        /* open the packet file for readback */
        OpenPcap();
    }

    /*
     * if daemon mode requested, fork daemon first, otherwise on linux
     * interface will be reset.
     */
    if(pv.daemon_flag)
    {
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Entering daemon mode\n"););
        GoDaemon();
    }

    if(pv.daemon_flag 
            || *pv.pidfile_suffix)
    {
        /* ... then create a PID file if not reading from a file */
        if (!pv.readmode_flag && (pv.daemon_flag || *pv.pidfile_suffix))
        {
#ifndef WIN32
            CreatePidFile(pv.interface);
#else
            CreatePidFile("WIN32");
#endif
        }
    }


    DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Setting Packet Processor\n"););

    /* set the packet processor (ethernet, slip, t/r, etc ) */
    SetPktProcessor();

#ifndef WIN32
    /* Drop the Chrooted Settings */
    if(pv.chroot_dir)
        SetChroot(pv.chroot_dir, &pv.log_dir);
    
    /* Drop privileges if requested, when initialization is done */
    SetUidGid();
    
#endif /*WIN32*/

    /* Tell 'em who wrote it, and what "it" is */
    DisplayBanner();

    if(pv.daemon_flag)
    {
        LogMessage("NETI@home initialization completed successfully\n");
    }

    DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Entering pcap loop\n"););

    InterfaceThread(NULL);

    return 0;
}

void ProcessPacket(char *user, struct pcap_pkthdr * pkthdr, u_char * pkt)
{
    Packet p;

	list<char *>::iterator leftover_Flows;
	list<int>::iterator leftover_sizes;
	int leftover_sizes_total = 0;
    char *tempemail;
    int netiver = -1;
#ifdef WIN32
    int oldprivacylevel;
#endif

    /* reset the packet flags for each packet */
    p.packet_flags = 0;

    pc.total++;

#if defined(WIN32) && defined(ENABLE_WIN32_SERVICE)
    if( pv.terminate_service_flag || pv.pause_service_flag )
    {
        return;
    }
#endif  /* WIN32 && ENABLE_WIN32_SERVICE */

    /* call the packet decoder */
    (*grinder) (&p, pkthdr, pkt);

	// If there is enough outstanding data or enough time has gone by and data exists, send it
	if((old_TCP_Flows.size() + old_UDP_Flows.size() + old_ICMP_Flows.size() + old_IGMP_Flows.size() > MAX_SIZE)
		|| ((!leftover_Flows_uncom.empty() || !old_TCP_Flows.empty() || !old_UDP_Flows.empty() || !old_ICMP_Flows.empty() || !old_IGMP_Flows.empty()) && (difftime(time(NULL), lastsendtime) > FORCE_SEND_TIME))) {
		send_stats();
#ifndef WIN32
        if(!updatealerted) {
            netiver = getLatestVersion();
            if(netiver == -1) {
                updatealerted = 0;
            }
            else if(netiver > NETI_VERSION) {
                ErrorMessage("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
                printf("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
                updatealerted = 1;
            }
            else {
                updatealerted = 1;
            }
        }
#endif
	}

	// If leftover sizes have gone over MAX_MSG, erase
	leftover_sizes = leftover_Flows_uncom_size.begin();
	while(leftover_sizes != leftover_Flows_uncom_size.end()) {
		leftover_sizes_total += *leftover_sizes;
		leftover_sizes++;
	}
	if(leftover_sizes_total >= MAX_MSG) {
		leftover_Flows = leftover_Flows_uncom.begin();
		while(leftover_Flows != leftover_Flows_uncom.end()) {
			free(*leftover_Flows);
			leftover_Flows_uncom.erase(leftover_Flows);
			leftover_Flows = leftover_Flows_uncom.begin();
		}
		leftover_Flows_uncom_size.clear();
        // Delete leftover Neti Header
        memset(&leftover_nh, 0, sizeof(leftover_nh));
        leftover_nh_valid = 0;
	}

    // Attempt to report email
    if(!pv.readmode_flag) {
        if((tempemail = getEmail()) != NULL) {
            if(lastemail == NULL) {
                if((difftime(time(NULL), lastemailsendtime) > FORCE_SEND_TIME)) {
                    send_email();
            	}
            }
            else {
                if((difftime(time(NULL), lastemailsendtime) > FORCE_SEND_TIME) && (strncmp(tempemail, lastemail, 256))) {
                    send_email();
            	}
            }
            free(tempemail);
        }
    }

#ifdef WIN32
    DWORD dwWaitResult;

    // Check Properties Event
    if(HProperties != NULL) {
        dwWaitResult = WaitForMultipleObjects(1, &HProperties, TRUE, 1);
        switch(dwWaitResult) {
            case WAIT_OBJECT_0:
                LogMessage("Properties Updated\n");
            	logfilesize = getLogfileSize();
            	if(logfilesize <= 0) {
                    ErrorMessage("Can't obtain Logfile Size from configuration file\n");
            		logfilesize = 1000000;
            	}
                oldprivacylevel = privacylevel;
            	privacylevel = getPrivacy();
            	if(privacylevel == -1) {
                    ErrorMessage("Can't obtain Privacy Setting from configuration file\n");
            		exit(1);
            	}
                if(oldprivacylevel != privacylevel) {
                    // Privacy Settings have been changed -- we should send the old stats first
                    privacylevel = oldprivacylevel;
                	if(!leftover_Flows_uncom.empty() || !old_TCP_Flows.empty() || !old_UDP_Flows.empty() || !old_ICMP_Flows.empty() || !old_IGMP_Flows.empty()) {
                		send_stats();
#ifndef WIN32
                        if(!updatealerted) {
                            netiver = getLatestVersion();
                            if(netiver == -1) {
                                updatealerted = 0;
                            }
                            else if(netiver > NETI_VERSION) {
                                ErrorMessage("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
                                printf("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
                                updatealerted = 1;
                            }
                            else {
                                updatealerted = 1;
                            }
                        }
#endif
                	}
                    privacylevel = getPrivacy();
                	if(privacylevel == -1) {
                        ErrorMessage("Can't obtain Privacy Setting from configuration file\n");
                		exit(1);
                	}
                }
                ResetEvent(HProperties);
                break;
            default:
                break;
        }
    }

    // Check Exit Event
    if(HExit != NULL) {
        dwWaitResult = WaitForMultipleObjects(1, &HExit, TRUE, 1);
        switch(dwWaitResult) {
            case WAIT_OBJECT_0:
                LogMessage("Exit selected\n");
                CleanExit(1);
                ResetEvent(HExit);
                break;
            default:
                break;
        }
    }
#endif
}


/*
 * Function: ShowUsage(char *)
 *
 * Purpose:  Display the program options and exit
 *
 * Arguments: progname => name of the program (argv[0])
 *
 * Returns: 0 => success
 */
int ShowUsage(char *progname)
{
    fprintf(stdout, "USAGE: %s [-options]\n", progname);
#if defined(WIN32) && defined(ENABLE_WIN32_SERVICE)
    fprintf(stdout, "       %s %s %s [-options]\n", progname
                                                                   , SERVICE_CMDLINE_PARAM
                                                                   , SERVICE_INSTALL_CMDLINE_PARAM);
    fprintf(stdout, "       %s %s %s\n", progname
                                       , SERVICE_CMDLINE_PARAM
                                       , SERVICE_UNINSTALL_CMDLINE_PARAM);
    fprintf(stdout, "       %s %s %s\n", progname
                                       , SERVICE_CMDLINE_PARAM
                                       , SERVICE_SHOW_CMDLINE_PARAM);
#endif

#ifdef WIN32
    #define FPUTS_WIN32(msg) fputs(msg,stdout)
    #define FPUTS_UNIX(msg)  NULL
    #define FPUTS_BOTH(msg)  fputs(msg,stdout)
#else
    #define FPUTS_WIN32(msg) 
    #define FPUTS_UNIX(msg)  fputs(msg,stdout)
    #define FPUTS_BOTH(msg)  fputs(msg,stdout)
#endif

    FPUTS_BOTH ("Options:\n");
    FPUTS_UNIX ("        -D         Run NETI@home in background (daemon) mode\n");
    FPUTS_UNIX ("        -g <gname> Run NETI@home gid as <gname> group (or gid) after initialization\n");
    FPUTS_BOTH ("        -i <if>    Listen on interface <if>\n");
    FPUTS_BOTH ("        -I <IP>    Specify local IP to be used when reading from file\n");
    FPUTS_BOTH ("        -l <ld>    Log to directory <ld>\n");
    fprintf(stdout, "        -P <snap>  Set explicit snaplen of packet (default: %d)\n",
                                    SNAPLEN);
    FPUTS_BOTH ("        -r <tf>    Read and process tcpdump file <tf> - must be used with -I\n");
    FPUTS_BOTH ("        -R <id>    Include 'id' in neti_intf<id>.pid file name\n");
    FPUTS_UNIX ("        -t <dir>   Chroots process to <dir> after initialization\n");
    FPUTS_UNIX ("        -u <uname> Run NETI@home uid as <uname> user (or uid) after initialization\n");
    FPUTS_BOTH ("        -V         Show version number\n");
    FPUTS_BOTH ("        -W         Lists available interfaces\n");
    FPUTS_BOTH ("        -?         Show this information\n");

#undef FPUTS_WIN32
#undef FPUTS_UNIX
#undef FPUTS_BOTH

    return 0;
}



/*
 * Function: ParseCmdLine(int, char *)
 *
 * Purpose:  Parse command line args
 *
 * Arguments: argc => count of arguments passed to the routine
 *            argv => 2-D character array, contains list of command line args
 *
 * Returns: 0 => success, 1 => exit on error
 */
extern char *optarg;                /* for getopt */
extern int   optind;                /* for getopt */

int ParseCmdLine(int argc, char *argv[])
{
    int ch;                         /* storage var for getopt info */
    char errorbuf[PCAP_ERRBUF_SIZE];
    int adaplen;
    char *valid_options;
    if_info_t *tempdev = NULL;
    int err;
    int devcount;
    int readip_set = 0;

    DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Parsing command line...\n"););

    /* just to be sane.. */
    username = NULL;
    groupname = NULL;
    pv.pidfile_suffix[0] = 0;

#ifndef WIN32
    /* Unix does not support an argument to -s <wink marty!> OR -E, (-W - no longer) */
    valid_options = "?:Dg:i:I:l:P:r:R:t:u:VW";
#else
    /* Win32 does not support:  -D, -g, -m, -t, -u */
    /* Win32 no longer supports an argument to -s, either! */
    valid_options = "?:i:I:l:P:r:R:VW";
#endif

    /* loop through each command line var and process it */
    while((ch = getopt(argc, argv, valid_options)) != -1)
    {
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Processing cmd line switch: %c\n", ch););
        switch(ch)
        {
            case 'D':                /* daemon mode */
#ifdef WIN32
                FatalError("Setting the Daemon mode is not supported in the "
                           "WIN32 port of NETI@home!  Use 'neti /SERVICE ...' "
                           "instead\n");
#endif
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Daemon mode flag set\n"););
                pv.daemon_flag = 1;
                break;

            case 'g':                /* setgid handler */
#ifdef WIN32
                FatalError("Setting the group id is not supported in the WIN32 port of NETI@home!\n");
#else
                if(groupname != NULL)
                    free(groupname);
                if((groupname = (char *)calloc(strlen(optarg) + 1, 1)) == NULL)
                    FatalPrintError("malloc");

                bcopy(optarg, groupname, strlen(optarg));

                if((groupid = atoi(groupname)) == 0)
                {
                    gr = getgrnam(groupname);
                    if(gr == NULL)
                        FatalError("Group \"%s\" unknown\n", groupname);

                    groupid = gr->gr_gid;
                }
#endif
                break;

            case 'i':
                if(pv.interface)
                {
                    FatalError("Cannot specify more than one network "
                               "interface on the command line.\n");
                }
                /* first, try to handle the "-i1" case, where an interface
                 * is specified by number.  If this fails, then fall-through
                 * to the case outside the ifdef/endif, where an interface
                 * can be specified by its fully qualified name, like as is
                 * shown by running 'neti -W', ie.
                 * "\Device\Packet_{12345678-90AB-CDEF-1234567890AB}"
                 */
                adaplen = atoi(optarg);
                if( adaplen > 0 )
                {
                    pv.interface_list = get_interface_list(&err, errorbuf);
                    if(pv.interface_list == NULL) {
                        perror(errorbuf);
                        exit(1);
                    }
                    tempdev = pv.interface_list;
                    devcount = 1;
                    while(tempdev != NULL && devcount != adaplen) {
                        tempdev = tempdev->next;
                        devcount++;
                    }
                    if(tempdev != NULL && devcount == adaplen) {
                        pv.interface = tempdev->name;
                    }
                    else {
                        pv.interface = NULL;
                    }

                    if ( pv.interface == NULL )
                    {
                        LogMessage("Invalid interface '%d'.", atoi(optarg));
                        exit(1);
                    }


                    DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Interface = %s\n",
                                PRINT_INTERFACE(pv.interface)));
                }
                else
                /* this code handles the case in which the user specifies
                   the entire name of the interface and it is compiled
                   regardless of which OS you have */
                {
                    if((pv.interface = (char *)malloc(strlen(optarg) + 1)) == NULL) {
                        FatalError("pv.interface malloc error\n");
                    }
                    /* XXX OOM check */
                    strlcpy(pv.interface, optarg, strlen(optarg)+1);
                    DEBUG_WRAP(DebugMessage(DEBUG_INIT,
                        "Interface = %s\n",
                        PRINT_INTERFACE(pv.interface)););
                }
                break;

            case 'I':
                if(inet_aton(optarg, &pv.readip) == 0) {
                    FatalError("error specifying readip\n");
                }
                readip_set = 1;
                break;

            case 'l':                /* use log dir <X> */
                if(!(pv.log_dir = strdup(optarg)))
                {
                    FatalError("Out of memory processing command line\n");
                }

                if(access(pv.log_dir, 2) != 0)
                {
                    FatalError("log directory '%s' does not exist\n", 
                            pv.log_dir);
                }
                break;

            case 'P':  /* explicitly define snaplength of packets */
                pv.pkt_snaplen = atoi(optarg);
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Snaplength of Packets set to: %d\n", pv.pkt_snaplen););
                break;

            case 'r':  /* read packets from a TCPdump file instead
                        * of the net */
                strlcpy(pv.readfile, optarg, STD_BUF);
                pv.readmode_flag = 1;
                break;

            case 'R': /* augment pid file name CPW*/
                if (strlen(optarg) < MAX_PIDFILE_SUFFIX && strlen(optarg) > 0)
                {
                    if (!strstr(optarg, "..") && !(strstr(optarg, "/")))
                    {
                        snprintf(pv.pidfile_suffix, MAX_PIDFILE_SUFFIX, "%s",
                                optarg);
                    }
                    else
                    {
                        FatalError("ERROR: illegal pidfile suffix: %s\n",
                                optarg);
                    }
                }
                else
                {
                    FatalError("ERROR: pidfile suffix length problem: %d\n",
                            strlen(optarg) );
                }
                break;

            case 't':  /* chroot to the user specified directory */
#ifdef WIN32
                FatalError("Setting the chroot directory is not supported in "
                           "the WIN32 port of NETI@home!\n");
#endif  /* WIN32 */
                if(!(pv.chroot_dir = strdup(optarg)))
                    FatalError("Out of memory processing command line\n");
                break;

            case 'u':  /* setuid */
#ifdef WIN32
                FatalError("Setting the user id is not "
                           "supported in the WIN32 port of NETI@home!\n");
#else
                if((username = (char *)calloc(strlen(optarg) + 1, 1)) == NULL)
                    FatalPrintError("malloc");

                bcopy(optarg, username, strlen(optarg));

                if((userid = atoi(username)) == 0)
                {
                    pw = getpwnam(username);
                    if(pw == NULL)
                        FatalError("User \"%s\" unknown\n", username);

                    userid = pw->pw_uid;
                }
                else
                {
                    pw = getpwuid(userid);
                    if(pw == NULL)
                        FatalError(
                                "Can not obtain username for uid: %lu\n",
                                (u_long) userid);
                }

                if(groupname == NULL)
                {
                    char name[256];

                    snprintf(name, 255, "%lu", (u_long) pw->pw_gid);

                    if((groupname = (char *)calloc(strlen(name) + 1, 1)) == NULL)
                    {
                        FatalPrintError("malloc");
                    }
                    groupid = pw->pw_gid;
                }
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "UserID: %lu GroupID: %lu\n",
                    (unsigned long) userid, (unsigned long) groupid););
#endif  /* !WIN32 */
                break;

            case 'V': /* prog ver already gets printed out, so we
                       * just exit */
                DisplayBanner();
                exit(0);

            case 'W':
                pv.interface_list = get_interface_list(&err, errorbuf);
                if(pv.interface_list == NULL) {
                    switch(err) {
                        case CANT_GET_INTERFACE_LIST:
                            printf("CANT_GET_INTERFACE_LIST: %s\n", errorbuf);
                            exit(0);
                            break;
                        case NO_INTERFACES_FOUND:
                            printf("NO_INTERFACES_FOUND: %s\n", errorbuf);
                            exit(0);
                            break;
                        default:
                            printf("Unknown Error: %d\t%s\n", err, errorbuf);
                            exit(0);
                            break;
                    }
                }
                tempdev = pv.interface_list;
                devcount = 1;
                DisplayBanner();
                printf("Interface\tDevice\tDescription\n");
                printf("------------------------------\n");
                while(tempdev != NULL) {
                    printf("%d\t", devcount);
                    if(tempdev->name) {
                        printf("%s\t", tempdev->name);
                    }
                    else {
                        printf("???\t");
                    }
                    if(tempdev->description) {
                        printf("%s\n", tempdev->description);
                    }
                    else {
                        printf("\n");
                    }
                    tempdev = tempdev->next;
                    devcount++;
                }
                printf("\n");

                exit(0);
                break;

            case '?':  /* show help and exit */
                DisplayBanner();
                ShowUsage(progname);
                exit(0);
        }
    }

    if(pv.readmode_flag && !readip_set) {
        FatalError("When reading from TCPdump files (-r), you must specify the \"local\" IP (-I)\n");
    }

    /* TODO relocate all of this to later in startup process */

    pv.pcap_cmd = NULL;        /* if NETI needs filters, set them here */

    if((pv.interface == NULL) && !pv.readmode_flag)
    {
        if(!(pv.interface = getInterface())) {
            pv.interface = pcap_lookupdev(errorbuf);
        }

        if(pv.interface == NULL)
            FatalError( "Failed to lookup for interface: %s."
                    " Please specify one with -i switch\n", errorbuf);
    }

    return 0;
}



/*
 * Function: SetPktProcessor()
 *
 * Purpose:  Set which packet processing function we're going to use based on
 *           what type of datalink layer we're using
 *
 * Arguments: int num => number of interface
 *
 * Returns: 0 => success
 */
int SetPktProcessor()
{
    switch(datalink)
    {
        case DLT_EN10MB:        /* Ethernet */
            grinder = DecodeEthPkt;
            break;

#ifdef DLT_IEEE802_11
        case DLT_IEEE802_11:
            grinder = DecodeIEEE80211Pkt;
            break;
#endif
        case 13:
        case DLT_IEEE802:                /* Token Ring */
            grinder = DecodeTRPkt;
            break;

        case DLT_FDDI:                /* FDDI */
            grinder = DecodeFDDIPkt;
            break;

#ifdef DLT_CHDLC
        case DLT_CHDLC:              /* Cisco HDLC */
            grinder = DecodeChdlcPkt;
            break;
#endif

        case DLT_SLIP:                /* Serial Line Internet Protocol */
            grinder = DecodeSlipPkt;
            break;

        case DLT_PPP:                /* point-to-point protocol */
            grinder = DecodePppPkt;
            break;

#ifdef DLT_PPP_SERIAL
        case DLT_PPP_SERIAL:         /* PPP with full HDLC header*/
            grinder = DecodePppSerialPkt;
            break;
#endif

#ifdef DLT_LINUX_SLL
        case DLT_LINUX_SLL:
            grinder = DecodeLinuxSLLPkt;
            break;
#endif

#ifdef DLT_PFLOG
        case DLT_PFLOG:
            grinder = DecodePflog;
            break;
#endif

#ifdef DLT_LOOP
        case DLT_LOOP:
#endif
        case DLT_NULL:            /* loopback and stuff.. you wouldn't run
                                   * NETI@home on it, but it's ok for
                                   * testing. */
            grinder = DecodeNullPkt;
            break;

#ifdef DLT_RAW /* Not supported in some arch or older pcap
                * versions */
        case DLT_RAW:
            grinder = DecodeRawPkt;
            break;
#endif
            /*
             * you need the I4L modified version of libpcap to get this stuff
             * working
             */
#ifdef DLT_I4L_RAWIP
        case DLT_I4L_RAWIP:
            grinder = DecodeI4LRawIPPkt;
            break;
#endif

#ifdef DLT_I4L_IP
        case DLT_I4L_IP:
            grinder = DecodeEthPkt;
            break;
#endif

#ifdef DLT_I4L_CISCOHDLC
        case DLT_I4L_CISCOHDLC:
            grinder = DecodeI4LCiscoIPPkt;
            break;
#endif

        default:                        /* oops, don't know how to handle this one */
            ErrorMessage("\n%s cannot handle data link type %d\n",
                    progname, datalink);
            CleanExit(1);
    }

    return 0;
}


/*
 * Function: void *InterfaceThread(void *arg)
 *
 * Purpose: wrapper for pthread_create() to create a thread per interface
 */
void *InterfaceThread(void *arg)
{
    static int intnum = 0;
    int myint;
    struct timeval starttime;
    struct timeval endtime;
    struct timeval difftime;
    struct timezone tz;

    myint = intnum;
    intnum++;

    bzero((char *) &tz, sizeof(tz));
    gettimeofday(&starttime, &tz);

    netistarttime = time(NULL);

    /* Read all packets on the device.  Continue until cnt packets read */
    if(pcap_loop(pd, -1, (pcap_handler) ProcessPacket, NULL) < 0)
    {
        if(pv.daemon_flag)
            syslog(LOG_CONS | LOG_DAEMON, "pcap_loop: %s", pcap_geterr(pd));
        else
            ErrorMessage("pcap_loop: %s\n", pcap_geterr(pd));

        CleanExit(1);
    }

    gettimeofday(&endtime, &tz);

    TIMERSUB(&endtime, &starttime, &difftime);

    printf("Run time for packet processing was %lu.%lu seconds\n", 
            difftime.tv_sec, difftime.tv_usec);

    CleanExit(0);

    return NULL;                /* avoid warnings */
}



/****************************************************************************
 *
 * Function: OpenPcap(char *, int)
 *
 * Purpose:  Open the libpcap interface
 *
 * Arguments: intf => name of the interface to open
 *            num  => number of the interface (to fill-in datalink and pd)
 *
 * Returns: 0 => success, exits on problems
 *
 ****************************************************************************/
int OpenPcap()
{
    bpf_u_int32 localnet, netmask;        /* net addr holders */
    struct bpf_program fcode;        /* Finite state machine holder */
    char errorbuf[PCAP_ERRBUF_SIZE];        /* buffer to put error strings in */
    int err;
    bpf_u_int32 defaultnet = 0xFFFFFF00;
    if_info_t *tempdev;

    /* if we're not reading packets from a file */
    if(pv.interface == NULL)
    {
        if (!pv.readmode_flag)
        {
            DEBUG_WRAP(DebugMessage(DEBUG_INIT,
                    "pv.interface is NULL, looking up interface....   "););
            /* look up the device and get the handle */
            if(!(pv.interface = getInterface())) {
                pv.interface = pcap_lookupdev(errorbuf);
            }
    
            DEBUG_WRAP(DebugMessage(DEBUG_INIT,
                    "found interface %s\n", PRINT_INTERFACE(pv.interface)););
            /* uh oh, we couldn't find the interface name */
            if(pv.interface == NULL)
            {
                FatalError("OpenPcap() interface lookup: \n\t%s\n",
               errorbuf);
            }
        }
        else
        {
            /* interface is null and we are in readmode */
            /* some routines would hate it to be NULL */
            pv.interface = "[reading from a file]"; 
        }
    }

    if (!pv.readmode_flag) {
        pv.interface_list = get_interface_list(&err, errorbuf);
        if(pv.interface_list == NULL) {
            switch(err) {
                case CANT_GET_INTERFACE_LIST:
                    FatalError("OpenPcap() CANT_GET_INTERFACE_LIST: %s\n", errorbuf);
                    break;
                case NO_INTERFACES_FOUND:
                    FatalError("OpenPcap() NO_INTERFACES_FOUND: %s\n", errorbuf);
                    break;
                default:
                    FatalError("OpenPcap() Unknown Error: %d\t%s\n", err, errorbuf);
                    break;
            }
        }
        if(strcasecmp(pv.interface, "any") != 0) {
            for(tempdev = pv.interface_list; tempdev != NULL; tempdev = tempdev->next) {
                if(tempdev->name != NULL) {
                    if(strcasecmp(tempdev->name, pv.interface) == 0) {
                        break;
                    }
                }
                if(tempdev->description != NULL) {
                    if(strcasecmp(tempdev->description, pv.interface) == 0) {
                        pv.interface = tempdev->name;
                        break;
                    }
                }
            }
            if(tempdev == NULL) {
                // End of list found without finding device!!!
                FatalError("OpenPcap() interface_list lookup\n");
            }
        }
        LogMessage("\nInitializing Network Interface %s\n", 
                PRINT_INTERFACE(pv.interface));
    }
    else {
        LogMessage("TCPDUMP file reading mode.\n");
    }

    if (!pv.readmode_flag)
    {
        if(pv.pkt_snaplen)        /* if it's set let's try it... */
        {
            if(pv.pkt_snaplen < MIN_SNAPLEN)        /* if it's < MIN set it to
                                                     * MIN */
            {
                /* XXX: Warning message, specified snaplen too small,
                 * snaplen set to X
                 */
                 snaplen = MIN_SNAPLEN;
            }
            else
            {
                 snaplen = pv.pkt_snaplen;
            }
         }
         else
         {
             snaplen = SNAPLEN;        /* otherwise let's put the compiled value in */
         }
        
        DEBUG_WRAP(DebugMessage(DEBUG_INIT,
                "snaplength info: set=%d/compiled=%d/wanted=%d\n",
                snaplen,  SNAPLEN, pv.pkt_snaplen););
    
        /* get the device file descriptor */
        pd = pcap_open_live(pv.interface, snaplen,
                0, READ_TIMEOUT, errorbuf);

    }
    else
    {   /* reading packets from a file */

        /* open the file */
        pd = pcap_open_offline(pv.readfile, errorbuf);

        /* the file didn't open correctly */
        if(pd == NULL)
        {
            FatalError("unable to open file \"%s\" for readback: %s\n",
                       pv.readfile, errorbuf);
        }
        /*
         * set the snaplen for the file (so we don't get a lot of extra crap
         * in the end of packets
         */
        snaplen = pcap_snapshot(pd);
    }

    /* something is wrong with the opened packet socket */
    if(pd == NULL)
    {
        if(strstr(errorbuf, "Permission denied"))
        {
            FatalError("You don't have permission to"
                       " sniff.\nTry doing this as root.\n");
        }
        else
        {
            FatalError("OpenPcap() device %s open: \n\t%s\n",
                       PRINT_INTERFACE(pv.interface), errorbuf);
        }
    }
    /* get local net and netmask */
    if(pcap_lookupnet(pv.interface, &localnet, &netmask, errorbuf) < 0)
    {
       if (!pv.readmode_flag)
       {
           ErrorMessage("OpenPcap() device %s network lookup: \n"
                        "\t%s\n",
                        PRINT_INTERFACE(pv.interface), errorbuf);

           if(pv.interface_list != NULL) {
               if(strcasecmp(pv.interface, "any") != 0) {
                   for(tempdev = pv.interface_list; tempdev != NULL; tempdev = tempdev->next) {
                       if(tempdev->name != NULL) {
                           if(strcasecmp(tempdev->name, pv.interface) == 0) {
                               break;
                           }
                       }
                       if(tempdev->description != NULL) {
                           if(strcasecmp(tempdev->description, pv.interface) == 0) {
                               pv.interface = tempdev->name;
                               break;
                           }
                       }
                   }
                   if(tempdev == NULL) {
                       // End of list found without finding device!!!
                       FatalError("netmask not found using interface_list\n");
                   }
                   else if(tempdev->addrs == NULL) {
                       FatalError("netmask not found using interface_list 2\n");
                   }
                   else {
                       netmask = tempdev->addrs->netmask.s_addr;
                   }
               }
               else {
                   // Not sure if I like this
                   netmask = htonl(defaultnet);
               }
           }
           else {
               FatalError("netmask lookup: interface_list == NULL\n");
           }
       }
       else {
           /*
            * set the default netmask to 255.255.255.0 (for stealthed
            * interfaces)
            */
           netmask = htonl(defaultnet);
       }
    }

    pv.netmask = netmask;

    /* compile BPF filter spec info fcode FSM */
    if(pcap_compile(pd, &fcode, pv.pcap_cmd, 1, netmask) < 0)
    {
        FatalError("OpenPcap() FSM compilation failed: \n\t%s\n"
                   "PCAP command: %s\n", pcap_geterr(pd), pv.pcap_cmd);
    }
    /* set the pcap filter */
    if(pcap_setfilter(pd, &fcode) < 0)
    {
        FatalError("OpenPcap() setfilter: \n\t%s\n",
                   pcap_geterr(pd));
    }
    
    /* get data link type */
    datalink = pcap_datalink(pd);

    if(datalink < 0)
    {
        FatalError("OpenPcap() datalink grab: \n\t%s\n",
                   pcap_geterr(pd));
    }
    return 0;
}

/* Signal Handlers ************************************************************/

static void SigTermHandler(int signal)
{
    CleanExit(0);
}

static void SigIntHandler(int signal)
{
    CleanExit(0);
}   

static void SigQuitHandler(int signal)
{
    CleanExit(0);
}

static void SigHupHandler(int signal)
{
    Restart();
}

static void SigUsr1Handler(int signal)
{
#ifndef WIN32
#if defined(LINUX) || defined(FREEBSD) || defined(OPENBSD) || defined(SOLARIS)
    sigset_t set;

    /* XXX why do we unblock all signals here? */
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
#else
    sigsetmask(0);
#endif
#endif  /* !WIN32 */

    // Report Stats
	if(!leftover_Flows_uncom.empty() || !old_TCP_Flows.empty() || !old_UDP_Flows.empty() || !old_ICMP_Flows.empty() || !old_IGMP_Flows.empty()) {
		send_stats();
	}
}

/**
 * dummy signal handler for nonroot users or chroot.
 *
 * @param signal signal to exec
 */
void SigCantHupHandler(int signal)
{
        LogMessage("Reload via Signal HUP does not work if you aren't root or are chroot'ed\n");
}

/****************************************************************************
 *
 * Function: CleanExit()
 *
 * Purpose:  Clean up misc file handles and such and exit
 *
 * Arguments: exit value;
 *
 * Returns: void function
 *
 ****************************************************************************/
void CleanExit(int exit_val)
{
    /* This function can be called more than once.  For example,
     * once from the SIGINT signal handler, and once recursively
     * as a result of calling pcap_close() below.  We only need
     * to perform the cleanup once, however.  So the static
     * variable already_exiting will act as a flag to prevent
     * double-freeing any memory.  Not guaranteed to be
     * thread-safe, but it will prevent the simple cases.
     */
    int netiver = -1;
    static int already_exiting = 0;
    if( already_exiting != 0 )
    {
        return;
    }
    already_exiting = 1;

    char *tempemail;

    /* free allocated memory */

	if(!leftover_Flows_uncom.empty() || !old_TCP_Flows.empty() || !old_UDP_Flows.empty() || !old_ICMP_Flows.empty() || !old_IGMP_Flows.empty()) {
		send_stats();
#ifndef WIN32
        if(!updatealerted) {
            netiver = getLatestVersion();
            if(netiver == -1) {
                updatealerted = 0;
            }
            else if(netiver > NETI_VERSION) {
                ErrorMessage("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
                printf("You are using an older version of NETI@home.  An update is available at http://neti.gatech.edu\n");
                updatealerted = 1;
            }
            else {
                updatealerted = 1;
            }
        }
#endif
	}

    // Attempt to report email
    if(!pv.readmode_flag) {
        if((tempemail = getEmail()) != NULL) {
            if(lastemail == NULL) {
                send_email();
            }
            else {
                if(strncmp(tempemail, lastemail, 256)) {
                    send_email();
            	}
            }
            free(tempemail);
        }
    }

    /* close pcap */
    if(pd) {
        pcap_close(pd);
    }

#ifdef WIN32
	closesocket(netimapfd);
	WSACleanup();
#else
	close(netimapfd);
#endif

    if(pv.interface_list) {
        free_interface_list(pv.interface_list);
        pv.interface_list = NULL;
    }

    LogMessage("NETI@home exiting\n");

    /* remove pid file */
    if(pv.pid_filename) {
        unlink(pv.pid_filename);
    }

    /* exit */
    exit(exit_val);
}

static void Restart()
{
    /* free allocated memory */

    /* close pcap */
    if(pd) {
        pcap_close(pd);
    }

    /* remove pid file */
    if(pv.pid_filename) {
        unlink(pv.pid_filename);
    }

    LogMessage("Restarting NETI@home\n");

    /* re-exec NETI@home */
#ifdef PARANOID
    execv(progname, progargs);
#else
    execvp(progname, progargs);
#endif

    /* only get here if we failed to restart */
    LogMessage("Restarting %s failed: %s", progname, strerror(errno));
    exit(1);
}

int send_stats() {
	int rc, n;
	char *buf;
	char *origbuf;
	char *tempbuf;
	char *bufcom;
	int origsize, bufsize;
	struct Neti_Header_v2 nh;
	uLongf comsize;
	map<struct Flow_Key, class TCP_Flow, fkcompare>::iterator tcpiterator;
	map<struct Flow_Key, class UDP_Flow, fkcompare>::iterator udpiterator;
	map<struct Flow_Key_NoPorts, class ICMP_Flow, fkcompare_noports>::iterator icmpiterator;
    list<class IGMP_Flow>::iterator igmpiterator;
	list<char *>::iterator leftover_Flows;
	list<int>::iterator leftover_sizes;
    char tempos[64];
    struct pcap_stat ps;
    int tempgeo = 0;
	struct hostent *he = NULL;

#ifdef WIN32
	SOCKET sockfd;
	SOCKADDR_IN remote_addr, my_addr;
#else
	int sockfd;
	struct sockaddr_in remote_addr, my_addr;
#endif

	origsize = (old_TCP_Flows.size() * sizeof(struct TCP_Stats_v2)) + (old_UDP_Flows.size() * sizeof(struct UDP_Stats_v2)) + (old_ICMP_Flows.size() * sizeof(struct ICMP_Stats_v2)) + (old_IGMP_Flows.size() * sizeof(struct IGMP_Stats_v2));
	leftover_sizes = leftover_Flows_uncom_size.begin();
	while(leftover_sizes != leftover_Flows_uncom_size.end()) {
		origsize += *leftover_sizes;
		leftover_sizes++;
	}
	bufsize = origsize;
	if((origbuf = (char *)malloc(bufsize)) == NULL) {
        ErrorMessage("malloc error\n");
		// For hysteresis
		lastsendtime += 3600;
		return -1;
	}
	comsize = bufsize + (bufsize / 10) + 12 + sizeof(struct Neti_Header_v2);
	if((bufcom = (char *)malloc(comsize)) == NULL) {
        ErrorMessage("bufcom malloc error\n");
		free(origbuf);
		// For hysteresis
		lastsendtime += 3600;
		return -1;
	}
	buf = origbuf;
	tempbuf = buf;

	// Add leftover_Flows_uncom to buffer and free and erase
	leftover_Flows = leftover_Flows_uncom.begin();
	leftover_sizes = leftover_Flows_uncom_size.begin();
	while(leftover_Flows != leftover_Flows_uncom.end()) {
		memcpy(buf, *leftover_Flows, *leftover_sizes);
		bufsize -= *leftover_sizes;
		buf += *leftover_sizes;
		free(*leftover_Flows);
		leftover_Flows_uncom.erase(leftover_Flows);
		leftover_Flows_uncom_size.erase(leftover_sizes);
		leftover_Flows = leftover_Flows_uncom.begin();
		leftover_sizes = leftover_Flows_uncom_size.begin();
	}

	// copy classes from old_UDP_Flows, old_TCP_Flows, old_ICMP_Flows, and old_IGMP_Flows into msg to send
	tcpiterator = old_TCP_Flows.begin();
	while(tcpiterator != old_TCP_Flows.end()) {
		tempbuf = buf;
		buf = tcpiterator->second.serialize(buf, &bufsize);
		if(buf == NULL) {
			buf = tempbuf;
			break;
		}
		old_TCP_Flows.erase(tcpiterator);
		tcpiterator = old_TCP_Flows.begin();
	}

	udpiterator = old_UDP_Flows.begin();
	while(udpiterator != old_UDP_Flows.end()) {
		tempbuf = buf;
		buf = udpiterator->second.serialize(buf, &bufsize);
		if(buf == NULL) {
			buf = tempbuf;
			break;
		}
		old_UDP_Flows.erase(udpiterator);
		udpiterator = old_UDP_Flows.begin();
	}

	icmpiterator = old_ICMP_Flows.begin();
	while(icmpiterator != old_ICMP_Flows.end()) {
		tempbuf = buf;
		buf = icmpiterator->second.serialize(buf, &bufsize);
		if(buf == NULL) {
			buf = tempbuf;
			break;
		}
		old_ICMP_Flows.erase(icmpiterator);
		icmpiterator = old_ICMP_Flows.begin();
	}

	igmpiterator = old_IGMP_Flows.begin();
	while(igmpiterator != old_IGMP_Flows.end()) {
		tempbuf = buf;
		buf = igmpiterator->serialize(buf, &bufsize);
		if(buf == NULL) {
			buf = tempbuf;
			break;
		}
		old_IGMP_Flows.erase(igmpiterator);
		igmpiterator = old_IGMP_Flows.begin();
	}


    // Set up socket to NETI@home Server
    if(!pv.readmode_flag) {
    	sockfd = socket(AF_INET, SOCK_STREAM, 0);

#ifdef WIN32
    	if(sockfd == INVALID_SOCKET) {
            ErrorMessage("Cannot open socket\n");
    		free(bufcom);
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}
#else
    	if(sockfd < 0) {
            ErrorMessage("Cannot open socket\n");
    		free(bufcom);
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}
#endif

    	my_addr.sin_family = AF_INET;
    	my_addr.sin_port = htons(0);
    	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    	memset(&(my_addr.sin_zero), '\0', 8);

    	remote_addr.sin_family = AF_INET;
    	remote_addr.sin_port = htons(DEST_PORT);

        // Get IP for DEST_ADDR
        he = gethostbyname(DEST_ADDR);
        if(he != NULL) {
#ifdef WIN32
            memcpy((char *)&remote_addr.sin_addr.S_un.S_addr, he->h_addr_list[0], he->h_length);
#else
            memcpy((char *)&remote_addr.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
#endif
    	}
    	else {
            ErrorMessage("Cannot obtain DEST_ADDR IP (send_stats)\n");
    		free(bufcom);
#ifdef WIN32
    		closesocket(sockfd);
#else
    		close(sockfd);
#endif
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}

    	memset(&(remote_addr.sin_zero), '\0', 8);

    	rc = bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr));
    	if(rc < 0) {
            ErrorMessage("Cannot bind port\n");
    		free(bufcom);
#ifdef WIN32
    		closesocket(sockfd);
#else
    		close(sockfd);
#endif
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}

    	rc = connect(sockfd, (struct sockaddr *)&remote_addr, sizeof(remote_addr));
    	if(rc < 0) {
            ErrorMessage("Cannot connect to server\n");
    		free(bufcom);
#ifdef WIN32
    		closesocket(sockfd);
#else
    		close(sockfd);
#endif
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}
    }
	
    // Create NETI@home Header
    
    if(!pv.readmode_flag && pd) {
        if(pcap_stats(pd, &ps)) {
            pcap_perror(pd, "pcap_stats");

            ErrorMessage("Cannot retrieve pcap stats\n");
    		free(bufcom);
#ifdef WIN32
    		closesocket(sockfd);
#else
    		close(sockfd);
#endif
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
        }
    }

	nh.magicnumber = htonl(NETI_MAGIC_NUMBER);          // 32 bits
	nh.netiversion = htonl(NETI_VERSION);               // 32 bits
    // Use leftover Neti Header, if it exists
    if(leftover_nh_valid) {
        nh.starttime = leftover_nh.starttime;           // 32 bits
    }
    else {
        nh.starttime = htonl(netistarttime);            // 32 bits
    }
	nh.uncomsize = htonl(origsize - bufsize);           // 32 bits
    tempgeo = getIdentifier();
    if(tempgeo < 0) {
        tempgeo = 0;
    }
    nh.identifier = htonl(tempgeo);                     // 32 bits
// XXX - What if leftover_nh privacylevel differs?
    nh.privacylevel = htons(privacylevel);              // 16 bits
    tempgeo = getGeolocation();
    if(tempgeo < 0) {
        tempgeo = 0;
    }
    nh.geolocation = htons(tempgeo);                    // 16 bits
    tempgeo = getUSZip();
    if(tempgeo < 0) {
        tempgeo = 0;
    }
    if(tempgeo > 99999) {
        tempgeo = 0;
    }
    nh.uszip = htonl(tempgeo);                          // 32 bits
    nh.datalink = htonl(datalink);                      // 32 bits

    // If reading from file, some things are different
    if(pv.readmode_flag) {
        // Use leftover Neti Header, if it exists
        if(leftover_nh_valid) {
            nh.totalpkts = htonl(ntohl(leftover_nh.totalpkts)
                + pc.tcp
                + pc.udp
                + pc.icmp
                + pc.igmp
                + pc.ipv6
                + pc.frags
                + pc.arp
                + pc.ipx
                + pc.eapol
                + pc.other
                + pc.discards);
        }
        else {
            nh.totalpkts = htonl(pc.tcp
                + pc.udp
                + pc.icmp
                + pc.igmp
                + pc.ipv6
                + pc.frags
                + pc.arp
                + pc.ipx
                + pc.eapol
                + pc.other
                + pc.discards);
        }
        nh.droppkts = htonl(0);
    }
    else {
        // Use leftover Neti Header, if it exists
        if(leftover_nh_valid) {
            nh.totalpkts = htonl(ntohl(leftover_nh.totalpkts) + ps.ps_recv + ps.ps_drop - lastps.ps_recv - lastps.ps_drop);
            nh.droppkts = htonl(ntohl(leftover_nh.droppkts) + ps.ps_drop - lastps.ps_drop);
        }
        else {
            nh.totalpkts = htonl(ps.ps_recv + ps.ps_drop - lastps.ps_recv - lastps.ps_drop);      // 32 bits
            nh.droppkts = htonl(ps.ps_drop - lastps.ps_drop);                    // 32 bits
        }

        // Since counters are not reset in pcap_stats
        lastps.ps_recv = ps.ps_recv;
        lastps.ps_drop = ps.ps_drop;
    }

    // Use leftover Neti Header, if it exists
    if(leftover_nh_valid) {
        nh.discpkts = htonl(ntohl(leftover_nh.discpkts) + pc.discards);
        nh.tcppkts = htonl(ntohl(leftover_nh.tcppkts) + pc.tcp);
        nh.udppkts = htonl(ntohl(leftover_nh.udppkts) + pc.udp);
        nh.icmppkts = htonl(ntohl(leftover_nh.icmppkts) + pc.icmp);
        nh.igmppkts = htonl(ntohl(leftover_nh.igmppkts) + pc.igmp);
        nh.ipv6pkts = htonl(ntohl(leftover_nh.ipv6pkts) + pc.ipv6);
        nh.fragpkts = htonl(ntohl(leftover_nh.fragpkts) + pc.frags);
        nh.arppkts = htonl(ntohl(leftover_nh.arppkts) + pc.arp);
        nh.ipxpkts = htonl(ntohl(leftover_nh.ipxpkts) + pc.ipx);
        nh.eapolpkts = htonl(ntohl(leftover_nh.eapolpkts) + pc.eapol);
        nh.otherpkts = htonl(ntohl(leftover_nh.otherpkts) + pc.other);
        nh.wifimgmtpkts = htonl(ntohl(leftover_nh.wifimgmtpkts) + pc.wifi_mgmt);
        nh.wificontpkts = htonl(ntohl(leftover_nh.wificontpkts) + pc.wifi_control);
        nh.wifidatapkts = htonl(ntohl(leftover_nh.wifidatapkts) + pc.wifi_data);
    }
    else {
        nh.discpkts = htonl(pc.discards);                   // 32 bits
        nh.tcppkts = htonl(pc.tcp);                         // 32 bits
        nh.udppkts = htonl(pc.udp);                         // 32 bits
        nh.icmppkts = htonl(pc.icmp);                       // 32 bits
        nh.igmppkts = htonl(pc.igmp);                       // 32 bits
        nh.ipv6pkts = htonl(pc.ipv6);                       // 32 bits
        nh.fragpkts = htonl(pc.frags);                      // 32 bits
        nh.arppkts = htonl(pc.arp);                         // 32 bits
        nh.ipxpkts = htonl(pc.ipx);                         // 32 bits
        nh.eapolpkts = htonl(pc.eapol);                     // 32 bits
        nh.otherpkts = htonl(pc.other);                     // 32 bits
        nh.wifimgmtpkts = htonl(pc.wifi_mgmt);              // 32 bits
        nh.wificontpkts = htonl(pc.wifi_control);           // 32 bits
        nh.wifidatapkts = htonl(pc.wifi_data);              // 32 bits
    }

    // Reset PacketCount (pc)
    memset(&pc, 0, sizeof(pc));

#ifdef RPM
    strncpy(nh.ostype, "Pre-Compiled RPM", (sizeof(nh.ostype) / sizeof(char)) - 1);
#else
#ifndef NETI_SYSTEM_TYPE
#ifdef WIN32
    strncpy(tempos, getWindowsSystemType().c_str(), (sizeof(tempos) / sizeof(char)) - 1);
    tempos[(sizeof(tempos) / sizeof(char)) - 1] = 0;
    strip(tempos);
    strncpy(nh.ostype, tempos, (sizeof(nh.ostype) / sizeof(char)) - 1);
#else
	strncpy(nh.ostype, "NETI_UNKNOWN_OS", (sizeof(nh.ostype) / sizeof(char)) - 1);
#endif
#else
    strncpy(tempos, NETI_SYSTEM_TYPE, (sizeof(tempos) / sizeof(char)) - 1);
    tempos[(sizeof(tempos) / sizeof(char)) - 1] = 0;
    strip(tempos);
    strncpy(nh.ostype, tempos, (sizeof(nh.ostype) / sizeof(char)) - 1);
#endif
#endif

	nh.ostype[(sizeof(nh.ostype) / sizeof(char)) - 1] = '\0';

	nh.sendtime = htonl(time(NULL));                    // 32 bits
    // End NETI@home Header Creation

    // Compress Data to Send to NETI@home Server and Send
    if(!pv.readmode_flag) {
    	memcpy(bufcom, &nh, sizeof(struct Neti_Header_v2));
    	bufcom += sizeof(struct Neti_Header_v2);
    	if(compress((Bytef *)bufcom, &comsize, (Bytef *)origbuf, (unsigned long)(origsize - bufsize)) != Z_OK) {
            ErrorMessage("Compression error\n");
    		free(bufcom - sizeof(struct Neti_Header_v2));
#ifdef WIN32
    		closesocket(sockfd);
#else
    		close(sockfd);
#endif
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
            // Create leftover Neti Header
            memcpy(&leftover_nh, &nh, sizeof(struct Neti_Header_v2));
            leftover_nh_valid = 1;
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}
    	bufcom -= sizeof(struct Neti_Header_v2);

    	n = send(sockfd, bufcom, comsize + sizeof(struct Neti_Header_v2), 0);
#ifdef WIN32
    	if(n == SOCKET_ERROR) {
            ErrorMessage("Cannot send data\n");
    		free(bufcom);
    		closesocket(sockfd);
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
            // Create leftover Neti Header
            memcpy(&leftover_nh, &nh, sizeof(struct Neti_Header_v2));
            leftover_nh_valid = 1;
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}
#else
    	if(n < 0) {
            ErrorMessage("Cannot send data\n");
    		free(bufcom);
    		close(sockfd);
    		// Add stuff back to old flows
    		leftover_Flows_uncom.push_back(origbuf);
    		leftover_Flows_uncom_size.push_back(origsize - bufsize);
            // Create leftover Neti Header
            memcpy(&leftover_nh, &nh, sizeof(struct Neti_Header_v2));
            leftover_nh_valid = 1;
    		// For hysteresis
    		lastsendtime += 3600;
    		return -1;
    	}
#endif
    }

	lastsendtime = time(NULL);
    // Reset after successful send
    netistarttime = time(NULL);

	// Write output to log file
    writeNETILog(origsize, bufsize, nh, origbuf);

	free(origbuf);
	free(bufcom);

    if(!pv.readmode_flag) {
#ifdef WIN32
    	closesocket(sockfd);
#else
    	close(sockfd);
#endif
    }
	return 0;
}

int send_email() {
	int rc, n;
	char *buf = NULL;
    struct Email_Header_v2 *emailhead = NULL;
    char *email = NULL;
	struct hostent *he = NULL;

#ifdef WIN32
	SOCKET sockfd;
	SOCKADDR_IN remote_addr, my_addr;
#else
	int sockfd;
	struct sockaddr_in remote_addr, my_addr;
#endif

    if(pv.readmode_flag) {
        return 0;
    }

    if((buf = (char *)malloc(sizeof(struct Email_Header_v2))) == NULL) {
        ErrorMessage("send_email: malloc error\n");
		// For hysteresis
		lastemailsendtime += 3600;
		return -1;
	}
    emailhead = (struct Email_Header_v2 *)buf;

    // Set up socket to NETI@home Server
   	sockfd = socket(AF_INET, SOCK_STREAM, 0);

#ifdef WIN32
   	if(sockfd == INVALID_SOCKET) {
        ErrorMessage("send_email: Cannot open socket\n");
        free(buf);
		// For hysteresis
		lastemailsendtime += 3600;
   		return -1;
   	}
#else
   	if(sockfd < 0) {
        ErrorMessage("send_email: Cannot open socket\n");
        free(buf);
		// For hysteresis
		lastemailsendtime += 3600;
   		return -1;
   	}
#endif

   	my_addr.sin_family = AF_INET;
   	my_addr.sin_port = htons(0);
   	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   	memset(&(my_addr.sin_zero), '\0', 8);

   	remote_addr.sin_family = AF_INET;
   	remote_addr.sin_port = htons(DEST_PORT);

	// Get IP for DEST_ADDR
	he = gethostbyname(DEST_ADDR);
	if(he != NULL) {
#ifdef WIN32
        memcpy((char *)&remote_addr.sin_addr.S_un.S_addr, he->h_addr_list[0], he->h_length);
#else
        memcpy((char *)&remote_addr.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
#endif
	}
	else {
        ErrorMessage("send_email: Cannot obtain IP for DEST_ADDR\n");
        free(buf);
		// For hysteresis
		lastemailsendtime += 3600;
#ifdef WIN32
   		closesocket(sockfd);
#else
   		close(sockfd);
#endif
   		return -1;
	}
   	memset(&(remote_addr.sin_zero), '\0', 8);

   	rc = bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr));
   	if(rc < 0) {
        ErrorMessage("send_email: Cannot bind port\n");
        free(buf);
		// For hysteresis
		lastemailsendtime += 3600;
#ifdef WIN32
   		closesocket(sockfd);
#else
   		close(sockfd);
#endif
   		return -1;
   	}

   	rc = connect(sockfd, (struct sockaddr *)&remote_addr, sizeof(remote_addr));
   	if(rc < 0) {
        ErrorMessage("send_email: Cannot connect to server\n");
        free(buf);
		// For hysteresis
		lastemailsendtime += 3600;
#ifdef WIN32
   		closesocket(sockfd);
#else
   		close(sockfd);
#endif
   		return -1;
   	}
	
    // Create Email Header
    emailhead->magicnumber = htonl(NETI_EMAIL_MAGIC_NUMBER);    // 32 bits
    emailhead->netiversion = htonl(NETI_VERSION);               // 32 bits
    if((email = getEmail()) == NULL) {
        ErrorMessage("send_email: Cannot retrieve email from conf file\n");
        free(buf);
		// For hysteresis
		lastemailsendtime += 3600;
#ifdef WIN32
   		closesocket(sockfd);
#else
   		close(sockfd);
#endif
   		return -1;
    }
    memset(&(emailhead->email), '\0', (sizeof(emailhead->email) / sizeof(char)));
    strncpy(emailhead->email, email, (sizeof(emailhead->email) / sizeof(char)) - 1);
    emailhead->email[(sizeof(emailhead->email) / sizeof(char)) - 1] = '\0';
    emailhead->sendtime = htonl(time(NULL));                    // 32 bits
    // End Email Header Creation

    // Send Email to NETI@home Server
   	n = send(sockfd, buf, sizeof(struct Email_Header_v2), 0);
#ifdef WIN32
   	if(n == SOCKET_ERROR) {
        ErrorMessage("send_email: Cannot send data\n");
        free(buf);
        free(email);
		// For hysteresis
		lastemailsendtime += 3600;
   		closesocket(sockfd);
   		return -1;
   	}
#else
   	if(n < 0) {
        ErrorMessage("send_email: Cannot send data\n");
        free(buf);
        free(email);
		// For hysteresis
		lastemailsendtime += 3600;
   		close(sockfd);
   		return -1;
   	}
#endif

	lastemailsendtime = time(NULL);
    if(lastemail != NULL) {
        free(lastemail);
    }
    lastemail = email;

	free(buf);

#ifdef WIN32
   	closesocket(sockfd);
#else
   	close(sockfd);
#endif
	return 0;
}

// This function writes the buffer buf to the LOGFILE
// and is called by send_stats()
// Returns:
// 0 == Successful
// -1 == ERROR
int writeNETILog(int origsize, int bufsize, struct Neti_Header_v2 nh, char *origbuf) {
	FILE *fpt;
	int filesize;
	//char *logbuf;
	int32_t lognewhost;
	bool loghost = true;
	bool logheader = true;

	lognewhost = htonl(NETI_NEW_USER);

#ifdef WIN32
	if((fpt = fopen(logfilename, "ab")) == NULL) {
        ErrorMessage("Cannot open log file\n");
        return 0;
	}
#else
	if((fpt = fopen(LOGFILE, "ab")) == NULL) {
        ErrorMessage("Cannot open log file\n");
        return 0;
	}
#endif

    if(pv.readmode_flag) {
    	fwrite(&lognewhost, sizeof(int32_t), 1, fpt);
		fwrite(&nh, sizeof(struct Neti_Header_v2), 1, fpt);
		fwrite(origbuf, sizeof(char), (origsize - bufsize), fpt);
		fclose(fpt);
    }
	else {
		fseek(fpt, 0, SEEK_END);	// Go to end of file
		filesize = ftell(fpt);		// Get size of file
		fseek(fpt, 0, SEEK_SET);	// Go to beginning of file
		if(filesize > logfilesize) {
			// File is already too big
			// How did that happen?
            ErrorMessage("Weird file size\n");
		}
		if((int)sizeof(int32_t) + (int)sizeof(struct Neti_Header_v2) + (origsize - bufsize) > logfilesize) {
			// Need to trim data and delete old logfile
            ErrorMessage("Need to trim data\n");

			// Delete old logfile
			fclose(fpt);
#ifdef WIN32
			remove(logfilename);
			if((fpt = fopen(logfilename, "ab")) == NULL) {
                ErrorMessage("Cannot open log file 2\n");
                return 0;
			}
#else
			remove(LOGFILE);
			if((fpt = fopen(LOGFILE, "ab")) == NULL) {
                ErrorMessage("Cannot open log file 2\n");
                return 0;
			}
#endif

			// Trim data
			loghost = false;
			if((int)sizeof(struct Neti_Header_v2) + (origsize - bufsize) > logfilesize) {
				logheader = false;
			}
			if((origsize - bufsize) > logfilesize) {
				origsize = bufsize + logfilesize;
			}
		}
		else if(filesize + (int)sizeof(int32_t) + (int)sizeof(struct Neti_Header_v2) + (origsize - bufsize) > logfilesize) {
			// Need to trim original file
            ErrorMessage("Need to trim original file\n");

            /*
             * Let's just delete for now
			// Set pointer to appropriate location
			fseek(fpt, logfilesize - sizeof(int32_t) - sizeof(struct Neti_Header_v2) - (origsize - bufsize), SEEK_END);
			logbuf = (char *)malloc(logfilesize - sizeof(int32_t) - sizeof(struct Neti_Header_v2) - (origsize - bufsize));
            if(logbuf == NULL) {
                ErrorMessage("logbuf malloc error\n");
                return 0;
            }
			fread(logbuf, 1, logfilesize - sizeof(int32_t) - sizeof(struct Neti_Header_v2) - (origsize - bufsize), fpt);
            */

			// Now delete original logfile and create it again with new data
			fclose(fpt);
#ifdef WIN32
			remove(logfilename);
			if((fpt = fopen(logfilename, "ab")) == NULL) {
                ErrorMessage("Cannot open log file 2\n");
                return 0;
			}
#else
			remove(LOGFILE);
			if((fpt = fopen(LOGFILE, "ab")) == NULL) {
                ErrorMessage("Cannot open log file 2\n");
                return 0;
			}
#endif
            /*
             * Let's just delete for now
			else {
				fwrite(logbuf, 1, logfilesize - sizeof(int32_t) - sizeof(struct Neti_Header_v2) - (origsize - bufsize), fpt);
			}

			// Now free the buffer
			free(logbuf);
             *
             */
		}
		if(loghost) {
			fwrite(&lognewhost, sizeof(int32_t), 1, fpt);
		}
		if(logheader) {
			fwrite(&nh, sizeof(struct Neti_Header_v2), 1, fpt);
		}
		// This is not right, should write the last logfilesize bytes instead of first, if(origsize - bufsize > logfilesize)
		fwrite(origbuf, sizeof(char), (origsize - bufsize), fpt);
		fclose(fpt);
	}
    return 0;
}

// This function retrieves the privacy setting from the conf file
// or returns LOW if reading a TCPdump file
// Returns:
// 0 == LOW privacy (collect everything)
// 1 == MEDIUM privacy (only collect the network portion of the local and remote IP)
// 2 == HIGH privacy (collect no IPs)
// -1 == ERROR
int getPrivacy() {
	char text[100];
	char *collector;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return 0;
    }
    else {
        // Open Configuration File
#ifdef WIN32
        if((conffile = fopen(conffilename, "r")) == NULL) {
            FatalError("getPrivacy(): Cannot open configuration file\n");
        }
#else
        if((conffile = fopen(CONFFILE, "r")) == NULL) {
            FatalError("getPrivacy(): Cannot open configuration file\n");
        }
#endif
    	memset(text, '\0', 100);
    	while(fgets(text, 100, conffile) != NULL) {
    		collector = strtok(text, " \t=");
    		if(collector != NULL) {
    			if(!strncasecmp(collector, "PRIVACY", 8)) {
    				collector = strtok(NULL, " \t=");
    				if(collector != NULL) {
    					if(!strncasecmp(collector, "LOW", 3)) {
                            fclose(conffile);
    						return 0;
    					}
    					else if(!strncasecmp(collector, "MEDIUM", 6)) {
                            fclose(conffile);
    						return 1;
    					}
    					else if(!strncasecmp(collector, "HIGH", 4)) {
                            fclose(conffile);
    						return 2;
    					}
    					else {
                            fclose(conffile);
    						return -1;
    					}
    				}
    			}
    		}
    	}
        fclose(conffile);
    	return -1;
    }
}

// This function retrieves the logfile size from the conf file
// Returns:
// logfile size
int getLogfileSize() {
	char text[100];
	char *collector;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return 0;
    }

    // Open Configuration File
#ifdef WIN32
    if((conffile = fopen(conffilename, "r")) == NULL) {
        FatalError("getLogfileSize(): Cannot open configuration file\n");
    }
#else
    if((conffile = fopen(CONFFILE, "r")) == NULL) {
        FatalError("getLogfileSize(): Cannot open configuration file\n");
    }
#endif
	memset(text, '\0', 100);
	while(fgets(text, 100, conffile) != NULL) {
		collector = strtok(text, " \t=");
		if(collector != NULL) {
			if(!strncasecmp(collector, "LOGFILESIZE", 11)) {
				collector = strtok(NULL, " \t=");
				if(collector != NULL) {
                    fclose(conffile);
					return atoi(collector);
				}
			}
		}
	}
    fclose(conffile);
	return 0;
}

// This function retrieves the interface name from the conf file
// Returns:
// interface name
char * getInterface() {
	char text[100];
	char *collector;
    char *toBeReturned = NULL;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return NULL;
    }

    // Open Configuration File
#ifdef WIN32
    if((conffile = fopen(conffilename, "r")) == NULL) {
        FatalError("getInterface(): Cannot open configuration file\n");
    }
#else
    if((conffile = fopen(CONFFILE, "r")) == NULL) {
        FatalError("getInterface(): Cannot open configuration file\n");
    }
#endif
	memset(text, '\0', 100);
	while(fgets(text, 100, conffile) != NULL) {
		collector = strtok(text, " \t=");
		if(collector != NULL) {
			if(!strncasecmp(collector, "INTERFACE", 9)) {
				collector = strtok(NULL, "=\"\r\n");
                while(collector != NULL) {
                    if(!isspace(collector[0])) {
                        if((toBeReturned = (char *)malloc(strlen(collector) + 1)) == NULL) {
                            return NULL;
                        }
                        memset(toBeReturned, '\0', strlen(collector) + 1);
                        strncpy(toBeReturned, collector, strlen(collector) + 1);
                        fclose(conffile);
                        return toBeReturned;
                    }
                    else {
				        collector = strtok(NULL, "=\"\r\n");
                    }
				}
			}
		}
	}
    fclose(conffile);
	return NULL;
}

// This function retrieves the geolocation from the conf file
// Returns:
// geographical location or -1 on Error
int getGeolocation() {
	char text[100];
	char *collector;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return -1;
    }

    // Open Configuration File
#ifdef WIN32
    if((conffile = fopen(conffilename, "r")) == NULL) {
        FatalError("getGeolocation(): Cannot open configuration file\n");
    }
#else
    if((conffile = fopen(CONFFILE, "r")) == NULL) {
        FatalError("getGeolocation(): Cannot open configuration file\n");
    }
#endif
	memset(text, '\0', 100);
	while(fgets(text, 100, conffile) != NULL) {
		collector = strtok(text, " \t=");
		if(collector != NULL) {
			if(!strncasecmp(collector, "LOCATION", 8)) {
				collector = strtok(NULL, " \t=");
				if(collector != NULL) {
                    fclose(conffile);
					return atoi(collector);
				}
			}
		}
	}
    fclose(conffile);
	return -1;
}

// This function retrieves the US Zip Code from the conf file
// Returns:
// US Zip Code or -1 on Error
int getUSZip() {
	char text[100];
	char *collector;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return -1;
    }

    // Open Configuration File
#ifdef WIN32
    if((conffile = fopen(conffilename, "r")) == NULL) {
        FatalError("getUSZip(): Cannot open configuration file\n");
    }
#else
    if((conffile = fopen(CONFFILE, "r")) == NULL) {
        FatalError("getUSZip(): Cannot open configuration file\n");
    }
#endif
	memset(text, '\0', 100);
	while(fgets(text, 100, conffile) != NULL) {
		collector = strtok(text, " \t=");
		if(collector != NULL) {
			if(!strncasecmp(collector, "USZIP", 5)) {
				collector = strtok(NULL, " \t=");
				if(collector != NULL) {
                    fclose(conffile);
					return atoi(collector);
				}
			}
		}
	}
    fclose(conffile);
	return -1;
}

// This function retrieves the email address from the conf file
// Returns:
// email address
char * getEmail() {
	char text[256];
	char *collector;
    char *toBeReturned = NULL;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return NULL;
    }

    // Open Configuration File
#ifdef WIN32
    if((conffile = fopen(conffilename, "r")) == NULL) {
        FatalError("getEmail(): Cannot open configuration file\n");
    }
#else
    if((conffile = fopen(CONFFILE, "r")) == NULL) {
        FatalError("getEmail(): Cannot open configuration file\n");
    }
#endif
	memset(text, '\0', 256);
	while(fgets(text, 256, conffile) != NULL) {
		collector = strtok(text, " \t=");
		if(collector != NULL) {
			if(!strncasecmp(collector, "EMAIL", 5)) {
				collector = strtok(NULL, " \t=\r\n");
				if(collector != NULL) {
                    if((toBeReturned = (char *)malloc(strlen(collector) + 1)) == NULL) {
                        return NULL;
                    }
                    memset(toBeReturned, '\0', strlen(collector) + 1);
                    strncpy(toBeReturned, collector, strlen(collector) + 1);
                    fclose(conffile);
                    return toBeReturned;
				}
			}
		}
	}
    fclose(conffile);
	return NULL;
}

// This function retrieves the unique identifier from the conf file
// or creates one if there is not one
// Returns:
// 32-bit unique identifier
int getIdentifier() {
	char text[100];
	char *collector;
    uint32_t randomnumber;
    struct timeval currtime;
    struct timezone tz;
    FILE *conffile = NULL;

    if(pv.readmode_flag) {
        return -1;
    }

    // Open Configuration File
#ifdef WIN32
    if((conffile = fopen(conffilename, "r")) == NULL) {
        FatalError("getIdentifier(): Cannot open configuration file\n");
    }
#else
    if((conffile = fopen(CONFFILE, "r")) == NULL) {
        FatalError("getIdentifier(): Cannot open configuration file\n");
    }
#endif
	memset(text, '\0', 100);
	while(fgets(text, 100, conffile) != NULL) {
		collector = strtok(text, " \t=");
		if(collector != NULL) {
			if(!strncasecmp(collector, "IDENTIFIER", 10)) {
				collector = strtok(NULL, " \t=");
				if(collector != NULL) {
                    fclose(conffile);
					return atoi(collector);
				}
			}
		}
	}

    // Identifier not found, must be created, placed in conf file, and returned

    // We actually use time to uniquely identify
    bzero((char *) &tz, sizeof(tz));
    gettimeofday(&currtime, &tz);

    randomnumber = currtime.tv_sec;
    //randomnumber = currtime.tv_sec << 8;
    ////currtime.tv_usec = currtime.tv_usec >> 26;
    //currtime.tv_usec = currtime.tv_usec & (0x000000FF);
    //randomnumber += currtime.tv_usec;

    fclose(conffile);
    // Open Configuration File for Appending
#ifdef WIN32
    if((conffile = fopen(conffilename, "a")) == NULL) {
        ErrorMessage("Cannot open configuration file to add identifier\n");
        return -1;
    }
#else
    if((conffile = fopen(CONFFILE, "a")) == NULL) {
        ErrorMessage("Cannot open configuration file to add identifier\n");
        return -1;
    }
#endif
    // Add standard comments
    fprintf(conffile, "# This is the unique identifier.\n");
    fprintf(conffile, "#     Please do not change as it will only harm research efforts.\n");
    // Add unique identifier
    fprintf(conffile, "IDENTIFIER = %u\n", randomnumber);
    fflush(conffile);
    fclose(conffile);
    return randomnumber;
}

#ifndef WIN32

// Function to Obtain the latest NETI@home version
int getLatestVersion() {
	int rc;
	char buf[4096];
    int bufsize = 4096;
	int sockfd;
    struct sockaddr_in server_addr;
	struct hostent *he = NULL;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if(sockfd < 0) {
        ErrorMessage("Cannot open socket\n");
		return -1;
	}

	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(UPDATEALERT_PORT);

	// Get IP for DEST_ADDR
	he = gethostbyname(DEST_ADDR);
	if(he != NULL) {
        memcpy((char *)&server_addr.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
	}
	else {
        ErrorMessage("Cannot obtain IP for DEST_ADDR for update alert\n");
		close(sockfd);
		return -1;
	}
	memset(&(server_addr.sin_zero), '\0', 8);

	rc = connect(sockfd, (struct sockaddr *)&server_addr, sizeof(server_addr));
	if(rc < 0) {
        ErrorMessage("Cannot connect to server for update alert\n");
		close(sockfd);
		return -1;
	}

	rc = recv(sockfd, buf, bufsize, 0);
    while(rc > 0) {
        if(rc < 0) {
            ErrorMessage("recv() failed for update alert\n");
            close(sockfd);
            return -1;
        }
        bufsize -= rc;
		rc = recv(sockfd, buf + ((4096 - bufsize) * sizeof(char)), bufsize, 0);
	}

	close(sockfd);
	
    if(rc < 0) {
        return -1;
    }

    if(buf == NULL) {
        return -1;
    }

	return atoi(buf);
}

#else

string getWindowsSystemType() {
	SYSTEM_INFO siSysInfo;
	OSVERSIONINFOEX osvi;
	BOOL bOsVersionInfoEx;
	int i;
	char temp[200];
	string type = "";

	// Hardware Info
	GetSystemInfo(&siSysInfo);
 
	i = sprintf(temp, "%u-", siSysInfo.dwOemId);
	type.append(temp, i);

	i = sprintf(temp, "%u-", siSysInfo.dwPageSize);
	type.append(temp, i);

	i = sprintf(temp, "%u-", siSysInfo.dwActiveProcessorMask);
	type.append(temp, i);

	i = sprintf(temp, "%u-", siSysInfo.dwNumberOfProcessors);
	type.append(temp, i);
 
	i = sprintf(temp, "%u-", siSysInfo.dwProcessorType);
	type.append(temp, i);

	i = sprintf(temp, "%u-", siSysInfo.wProcessorLevel);
	type.append(temp, i);

	i = sprintf(temp, "%u-", siSysInfo.wProcessorRevision);
	type.append(temp, i);


   // Try calling GetVersionEx using the OSVERSIONINFOEX structure.
   //
   // If that fails, try using the OSVERSIONINFO structure.

   ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
   osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

   if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi)) )
   {
      // If OSVERSIONINFOEX doesn't work, try OSVERSIONINFO.

      osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
      if (! GetVersionEx ( (OSVERSIONINFO *) &osvi) ) 
         return type;
   }

   switch (osvi.dwPlatformId)
   {
      case VER_PLATFORM_WIN32_NT:
		  if ( osvi.dwMajorVersion <= 4 ) {
			type += "winNT";
		  }

		  if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 ) {
			type += "win2000";
		  }

		  if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1 ) {
			type += "winXP";
		  }

         if( bOsVersionInfoEx )
         {
            if ( osvi.wProductType == VER_NT_WORKSTATION )
            {
				if( osvi.wSuiteMask & VER_SUITE_PERSONAL ) {
				  type += "-per";
				}
				else {
				  type += "-pro";
				}
            }

            else if ( osvi.wProductType == VER_NT_SERVER )
            {
				if( osvi.wSuiteMask & VER_SUITE_DATACENTER ) {
				  type += "-dtasrv";
				}
				else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE ) {
				  type += "-advsrv";
				}
				else {
				  type += "-srv";
				}
            }
         }
         else
         {
            HKEY hKey;
            char szProductType[80];
            DWORD dwBufLen;

            RegOpenKeyEx( HKEY_LOCAL_MACHINE,
               "SYSTEM\\CurrentControlSet\\Control\\ProductOptions",
               0, KEY_QUERY_VALUE, &hKey );
            RegQueryValueEx( hKey, "ProductType", NULL, NULL,
               (LPBYTE) szProductType, &dwBufLen);
            RegCloseKey( hKey );
			if ( lstrcmpi( "WINNT", szProductType) == 0 ) {
			   type += "-pro";
			}
			if ( lstrcmpi( "LANMANNT", szProductType) == 0 ) {
			   type += "-srv";
			}
			if ( lstrcmpi( "SERVERNT", szProductType) == 0 ) {
			   type += "-advsrv";
			}
         }

         if ( osvi.dwMajorVersion <= 4 )
         {
	/*
            printf ("version %d.%d %s (Build %d)\n",
               osvi.dwMajorVersion,
               osvi.dwMinorVersion,
               osvi.szCSDVersion,
			   osvi.dwBuildNumber & 0xFFFF);
	*/

				i = sprintf(temp, "-%d", osvi.dwMajorVersion);
				type.append(temp, i);

				i = sprintf(temp, "-%d", osvi.dwMinorVersion);
				type.append(temp, i);

				i = sprintf(temp, "-%s", osvi.szCSDVersion);
				type.append(temp, i);

				i = sprintf(temp, "-%d", osvi.dwBuildNumber & 0xFFFF);
				type.append(temp, i);
		 }
         else
         { 
	/*
            printf ("%s (Build %d)\n",
               osvi.szCSDVersion,
               osvi.dwBuildNumber & 0xFFFF);
	*/

				i = sprintf(temp, "-%s", osvi.szCSDVersion);
				type.append(temp, i);

				i = sprintf(temp, "-%d", osvi.dwBuildNumber & 0xFFFF);
				type.append(temp, i);
         }
         break;

      case VER_PLATFORM_WIN32_WINDOWS:

         if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 0)
         {
			 type += "win95";
			 if ( osvi.szCSDVersion[1] == 'C' || osvi.szCSDVersion[1] == 'B' ) {
				type += "osr2";
				}
         } 

         if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 10)
         {
			 type += "win98";
			 if ( osvi.szCSDVersion[1] == 'A' ) {
				type += "SE";
				}
         } 

         if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 90)
         {
			 type += "winME";
         } 
         break;

      case VER_PLATFORM_WIN32s:

		 type += "win32s";
         break;
   }
   return type;
}

#endif


/* vim: smartindent:expandtab:sw=4:ts=4:tw=0
 */
