/* Copyright 2003-2005 Charles Robert Simpson, Jr.
 *
 * This file is part of NETI@home.
 * 
 * NETI@home is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * NETI@home is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NETI@home; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *
 * Program: devselect
 *
 * Purpose: Display interfaces for installer
 *
 * Author: Charles Robert Simpson, Jr. (Robby) (neti@ece.gatech.edu)
 *
 * Comments: Much of this code was taken from Martin Roesch's Snort program.
 *           Gotta love open source and code reuse!  See Martin's comments
 *           below:
 * Comments: Ideas and code stolen liberally from Mike Borella's IP Grab
 *           program. Check out his stuff at http://www.borella.net.  I
 *           also have ripped some util functions from TCPdump, plus Mike's
 *           prog is derived from it as well.  All hail TCPdump....
 *
 */

#include <stdio.h>
#include <pcap.h>
#include <winsock2.h>

#define MAX_WIN_IF_NAME_LEN 511
#define CANT_GET_INTERFACE_LIST 0
#define NO_INTERFACES_FOUND 1   /* list is empty */

struct _if_info_t;
struct _if_addr_t;

typedef struct _if_addr_t {
    struct in_addr ip;
    struct in_addr netmask;
    struct _if_addr_t *next;
} if_addr_t;

typedef struct _if_info_t {
    char *name;
    char *description;
    if_addr_t *addrs;
    struct _if_info_t *next;
} if_info_t;

void CleanExit(int);
int init_winsock(void);
if_info_t * get_interface_list(int *err, char *err_str);
if_info_t * get_interface_list_findalldevs(int *err, char *err_str);

/*
 *
 * Function: main(int, char *)
 *
 * Returns: 0 => normal exit, 1 => exit on error
 *
 */
int main(int argc, char* argv[]) 
{
    char errorbuf[PCAP_ERRBUF_SIZE];
    int err;
    int devcount;
    if_info_t *interface_list = NULL;
    FILE *fpt = NULL;

    if (!init_winsock())
        exit(1);

    if(argc < 2) {
        printf("Error, please specify output file on command line\n");
        exit(1);
    }

    if((fpt = fopen(argv[1], "w")) == NULL) {
        printf("Cannot open file %s\n", argv[1]);
        exit(1);
    }

    interface_list = get_interface_list(&err, errorbuf);
    if(interface_list == NULL) {
        CleanExit(0);
    }
    devcount = 0;
    while(interface_list != NULL) {
        if(interface_list->description) {
            if(devcount) {
                fprintf(fpt, "|");
            }
            fprintf(fpt, "%s", interface_list->description);
            devcount++;
        }
        else if(interface_list->name) {
            if(devcount) {
                fprintf(fpt, "|");
            }
            fprintf(fpt, "%s", interface_list->name);
            devcount++;
        }
        interface_list = interface_list->next;
    }
    fprintf(fpt, "\n");

    fclose(fpt);

    CleanExit(0);
}

/****************************************************************************
 *
 * Function: CleanExit()
 *
 * Purpose:  Clean up misc file handles and such and exit
 *
 * Arguments: exit value;
 *
 * Returns: void function
 *
 ****************************************************************************/
void CleanExit(int exit_val)
{
    /* This function can be called more than once.  For example,
     * once from the SIGINT signal handler, and once recursively
     * as a result of calling pcap_close() below.  We only need
     * to perform the cleanup once, however.  So the static
     * variable already_exiting will act as a flag to prevent
     * double-freeing any memory.  Not guaranteed to be
     * thread-safe, but it will prevent the simple cases.
     */
    static int already_exiting = 0;
    if( already_exiting != 0 )
    {
        return;
    }
    already_exiting = 1;

	WSACleanup();

    /* exit */
    exit(exit_val);
}

/****************************************************************************
 *
 * Function: init_winsock(void)
 *
 * Purpose:  Initialize winsock.
 *
 * Arguments: None.
 *
 * Returns: 0 => Initilization failed.
 *          1 => Initilization succeeded.
 *
 ****************************************************************************/
int init_winsock(void)
{
    WORD wVersionRequested = MAKEWORD(1, 1);
    WSADATA wsaData;

    if (WSAStartup(wVersionRequested, &wsaData))
    {
        return 0;
    }

    if (LOBYTE(wsaData.wVersion) < 1 || HIBYTE(wsaData.wVersion) < 1)
    {
        WSACleanup();
        return 0;
    }

    return 1;
}

// The following code was taken from Ethereal 0.10.4, with minor modification
if_info_t * if_info_new(char *name, char *description)
{
	if_info_t *if_info;

	if((if_info = (if_info_t *)malloc(sizeof (if_info_t))) == NULL) {
        return NULL;
    }
	if_info->name = strdup(name);
	if (description == NULL) {
		if_info->description = NULL;
	}
	else {
		if_info->description = strdup(description);
	}
	if_info->addrs = NULL;
	if_info->next = NULL;

	return if_info;
}

/*
 * This will use "pcap_findalldevs()" if we have it, otherwise it'll
 * fail.
 */
if_info_t * get_interface_list(int *err, char *err_str)
{
	return get_interface_list_findalldevs(err, err_str);
}

if_info_t * get_interface_list_findalldevs(int *err, char *err_str)
{
	if_info_t *il = NULL;
	if_info_t *count = NULL;
	pcap_if_t *alldevs, *dev;
	pcap_addr_t *addr;
	if_info_t *if_info;
	if_addr_t *if_addr = NULL;

	if (pcap_findalldevs(&alldevs, err_str) == -1) {
		*err = CANT_GET_INTERFACE_LIST;
		return NULL;
	}

	if (alldevs == NULL) {
		/*
		 * No interfaces found.
		 */
		*err = NO_INTERFACES_FOUND;
		return NULL;
	}

	for (dev = alldevs; dev != NULL; dev = dev->next) {
		if_info = if_info_new(dev->name, dev->description);
		for(addr = dev->addresses; addr != NULL; addr = addr->next) {
			if(addr->addr->sa_family == AF_INET) {
				if_addr = if_info->addrs;
				if(if_addr == NULL) {
					if_info->addrs = (if_addr_t *)malloc(sizeof(if_addr_t));
					if_addr = if_info->addrs;
				}
				else {
					while(if_addr->next != NULL) {
						if_addr = if_addr->next;
					}
					if_addr->next = (if_addr_t *)malloc(sizeof(if_addr_t));
					if_addr = if_addr->next;
				}
				memcpy((struct in_addr *)&if_addr->ip, (struct in_addr *)&((struct sockaddr_in *)addr->addr)->sin_addr, sizeof(struct in_addr));
				if(addr->netmask != NULL) {
					memcpy((struct in_addr *)&if_addr->netmask, (struct in_addr *)&((struct sockaddr_in *)addr->netmask)->sin_addr, sizeof(struct in_addr));
				}
				if_addr->next = NULL;
			}
		}
		if(il == NULL) {
			il = if_info;
		}
		else {
			for(count = il; count != NULL; count = count->next) {
				if(count->next == NULL) {
					count->next = if_info;
					break;
				}
			}
		}
	}
	pcap_freealldevs(alldevs);

	return il;
}

/* vim: smartindent:expandtab:sw=4:ts=4:tw=0
 */
